//****************************************************************************
// File: mainwnd.c
//
// Purpose: Main overlapped window's message handler
//
// Functions:
//    MainWndProc() - message handler for main overlapped window
//    DoCommands() - called by MainWndProc to handle all WM_COMMAND messages
//
// Development Team:
//
//
// Written by Microsoft Product Support Services, Windows Developer Support
// Copyright (c) 1992 Microsoft Corporation. All rights reserved.
//****************************************************************************

#include <windows.h>
#include <ctype.h>
#include <commdlg.h>
#include <memory.h>
#include "global.h"
#include "about.h"
#include "mainwnd.h"
#include "print.h"

// Local function prototypes.
long DoCommands (HWND hWnd,
                 unsigned message,
                 WORD wParam,
                 LONG lParam);

//****************************************************************************
// Function: MainWndProc
//
// Purpose: Message handler for main overlapped window.
//
// Parameters:
//    hWnd    == Handle to _this_ window.
//    message == Message to process.
//    wParam  == WORD parameter -- depends on message
//    lParam  == LONG parameter -- depends on message
//
// Returns: Depends on message.
//
// Comments:
//
// History:  Date       Author        Reason
//           1/27/92                  Created
//****************************************************************************

long FAR PASCAL MainWndProc (HWND hWnd,
                             unsigned message,
                             WORD wParam,
                             LONG lParam)
{
	switch (message)
	{
		case WM_CREATE:
		{
			HFILE hfText;
         
			// assumes that the file textfile.txt exists
			hfText = _lopen((LPCSTR)"textfile.txt", READ);
			glFileSize = _llseek(hfText, 0L, 2);
			_llseek(hfText, 0L, 0);
         
			// assumes that the memory is allocated and locked
			ghMem = GlobalAlloc(GHND, (DWORD)(glFileSize + 1L));
			glpszText = GlobalLock(ghMem);
        	 
			ghLineMem = GlobalAlloc(GHND, (DWORD)(MAXLINES*sizeof(LINEINFO)));
			glpLines = (LINEINFO FAR*)GlobalLock(ghLineMem);
         
			ghPageMem = GlobalAlloc(GHND, (DWORD)(MAXPAGES*sizeof(PAGEINFO)));
			glpPages = (PAGEINFO FAR*)GlobalLock(ghPageMem);
         
			// store the actual number of bytes read in guBuffLen
			guBuffLen = _lread(hfText, glpszText, (UINT)glFileSize);
			_lclose (hfText);
         
			// Use this font to format and paint text
			ghScrDC = GetDC(hWnd);         
			SetBkMode(ghScrDC, TRANSPARENT);
			gnScrXRes = GetDeviceCaps(ghScrDC, LOGPIXELSX);
			gnScrYRes = GetDeviceCaps(ghScrDC, LOGPIXELSY);
			
			// assume that a default printer is available to the system        
			ghPrnDC = GetDefaultPrinterDC(hWnd);
			SetBkMode(ghPrnDC, TRANSPARENT);		
			gnPrnXRes = GetDeviceCaps(ghPrnDC, LOGPIXELSX);
			gnPrnYRes = GetDeviceCaps(ghPrnDC, LOGPIXELSY);
			
			// set up screen DC for scaling         
			SetMapMode(ghScrDC, MM_ANISOTROPIC);
			SetWindowExt(ghScrDC, gnPrnXRes, gnPrnYRes);
			SetViewportExt(ghScrDC, gnScrXRes, gnScrYRes);
			grc.right = GetDeviceCaps(ghPrnDC, HORZRES);
			grc.bottom = GetDeviceCaps(ghPrnDC, VERTRES);
			
			// create a default font 
			glf.lfHeight = -MulDiv(12, gnPrnYRes, 72); 
			glf.lfOutPrecision = OUT_TT_PRECIS;
			lstrcpy(glf.lfFaceName, "Arial");
			ghFont = CreateFontIndirect((LPLOGFONT)&glf);
			
			// Remember the ChooseFont bug that relies on screen metrics to compute point size when 
			// it initializes the dialog based on the inbound LOGFONT. Work around that bug.
			glf.lfHeight = -MulDiv(12, gnScrYRes, 72);			

			SelectObject(ghScrDC, (HGDIOBJ)ghFont);
			SelectObject(ghPrnDC, (HGDIOBJ)ghFont);            
			FormatText(hWnd, ghScrDC, ghPrnDC);
			break;
		}
      
		case WM_SIZE:
		{
			TEXTMETRIC tm;
			UINT uLineHt;
			RECT rc;
			
			GetTextMetrics(ghPrnDC, (LPTEXTMETRIC)&tm);
			uLineHt = tm.tmHeight + tm.tmExternalLeading;
			guNumLinesPerPage = grc.bottom / uLineHt;
    
			GetClientRect(hWnd, (LPRECT)&rc);    
			guNumLinesPerScrPage = rc.bottom / MulDiv(uLineHt, gnScrXRes, gnPrnXRes);
			guLastVisLine = guFirstVisLine + guNumLinesPerScrPage - 1;
			if (guLastVisLine > guNumLines)
				guLastVisLine = guNumLines;
    		break;
    	}

		case WM_VSCROLL:
			switch (wParam)
			{
				case SB_LINEDOWN:
					guFirstVisLine++;
					Scroll(hWnd);
					break;
      			
				case SB_PAGEUP:
					guFirstVisLine -= guNumLinesPerScrPage;
					Scroll(hWnd);
					break;
      			
				case SB_PAGEDOWN:
					guFirstVisLine += guNumLinesPerScrPage;
					Scroll(hWnd);
					break;
      			
				case SB_LINEUP:
 					guFirstVisLine--;      		
					Scroll(hWnd);
					break;
      			
				case SB_THUMBPOSITION:
					break;
      			
				case SB_THUMBTRACK:
					break;
			}

			break;      
      
		case WM_PAINT:
		{
			PAINTSTRUCT ps;
			char szText[] = "No text is available. Either textfile.txt does not exist or is too large";
      	 
			BeginPaint(hWnd, (LPPAINTSTRUCT)&ps);
			SetBkMode(ps.hdc, TRANSPARENT);
			gnPrnXRes = GetDeviceCaps(ghPrnDC, LOGPIXELSX);
			gnPrnYRes = GetDeviceCaps(ghPrnDC, LOGPIXELSY);         
			SetMapMode(ps.hdc, MM_ANISOTROPIC);
			SetWindowExt(ps.hdc, gnPrnXRes, gnPrnYRes);
			SetViewportExt(ps.hdc, gnScrXRes, gnScrYRes);
	
			SelectObject(ps.hdc, (HGDIOBJ)ghFont);
 			if (guBuffLen <= 0)
				TextOut(ps.hdc, 0, 0, szText, lstrlen(szText));
			else
				// do the painting here.
				RepaintClient(ps.hdc, ghPrnDC, guFirstVisLine, guLastVisLine);				
			EndPaint(hWnd, (LPPAINTSTRUCT)&ps);
		}
         
      // Dispatch WM_COMMAND messages to our command handler, DoCommands().
      case WM_COMMAND:
         return DoCommands(hWnd, message, wParam, lParam);


      // On WM_DESTROY, terminate this app by posting a WM_QUIT message.

	case WM_DESTROY:
		// unlock and free the memory
		GlobalUnlock(ghMem);
		GlobalFree(ghMem);
		GlobalUnlock(ghLineMem);
		GlobalFree(ghLineMem);
		GlobalUnlock(ghPageMem);
		GlobalFree(ghPageMem);
	         
		SelectObject(ghScrDC, GetStockObject(SYSTEM_FONT));
		ReleaseDC(hWnd, ghScrDC);
		if (ghPrnDC)
		{
			SelectObject(ghPrnDC, GetStockObject(SYSTEM_FONT));
         	DeleteDC(ghPrnDC);                  
		}
		PostQuitMessage(0);
		break;


		// We didn't handle, pass to DefWindowProc.
	
	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
   }


   return (NULL);
}

//****************************************************************************
// Function: DoCommands
//
// Purpose: Called by MainWndProc() to handle all WM_COMMAND messages.
//
// Parameters:
//    hWnd    == Handle to _this_ window.
//    message == Message to process.
//    wParam  == WORD parameter -- depends on message
//    lParam  == LONG parameter -- depends on message
//
// Returns : Depends on message.
//
// Comments:
//
// History:  Date       Author        Reason
//           1/27/92                  Created
//****************************************************************************

long DoCommands (HWND hWnd,
                 unsigned message,
                 WORD wParam,
                 LONG lParam)
{
   switch (wParam)
   {
   
		case IDM_FONT:
		{
			CHOOSEFONT cf;
			HFONT hOldFont;

			// initialize the choosefont common dialog
			_fmemset((LPCHOOSEFONT)&cf, 0, sizeof(CHOOSEFONT));
			cf.lStructSize = sizeof(CHOOSEFONT);
			cf.hwndOwner = hWnd;
			cf.lpLogFont = (LPLOGFONT)&glf;
			cf.Flags = CF_INITTOLOGFONTSTRUCT | CF_BOTH | CF_SCALABLEONLY | CF_WYSIWYG | CF_LIMITSIZE;
			cf.hDC = ghPrnDC;
			cf.nSizeMax = 28;
         
			// Call the choosefont common dialog
			if (ChooseFont((LPCHOOSEFONT)&cf))
			{
				// Replace the old font with the new font             
				// account for the ChooseFont bug that computes logfont metrics with screen DC's
				// metrics even when the hDC member is initialized to a printer DC.
				(cf.lpLogFont)->lfHeight = -MulDiv(cf.iPointSize/10, gnPrnYRes, 72);
         		ghFont = CreateFontIndirect(cf.lpLogFont);
         		hOldFont = (HFONT)SelectObject(ghScrDC, (HGDIOBJ)ghFont);
         		SelectObject(ghPrnDC, (HGDIOBJ)ghFont);
         		DeleteObject(hOldFont);
         
         		// Remember the ChooseFont bug that relies on screen metrics to compute point size when 
         		// it initializes the dialog based on the inbound LOGFONT. Work around that bug.
         		(cf.lpLogFont)->lfHeight = -MulDiv(cf.iPointSize/10, gnScrYRes, 72);         
         
         		// force the scroll thumb to go to starting position and
         		// repaint text starting from the beginning
         		SetScrollPos(hWnd, SB_VERT, 0, TRUE);
         		FormatText(hWnd, ghScrDC, ghPrnDC);
         		InvalidateRect(hWnd, NULL, TRUE);
         	}
         	break;
		}
       
		case IDM_PRINT:
			PrintHandler(hWnd, ghPrnDC);
			break;
			
		case IDM_PREVIEW:
			PrintPreviewHandler(hWnd, ghPrnDC);
			break;	
       
		// Put up the About box.
		case IDM_ABOUT:
		{
			FARPROC lpDlgProc;

			lpDlgProc = MakeProcInstance(AboutDlg, ghInst);
			DialogBox(ghInst, AboutBoxName, hWnd, lpDlgProc);
			FreeProcInstance(lpDlgProc);
			break;
		}


      // User picked File.Exit so post WM_CLOSE, causing the default
      // window procedure to call DestroyWindow(). We could process
      // WM_CLOSE for confirmation from user that we should exit.

      case IDM_EXIT:
         PostMessage(hWnd, WM_CLOSE, 0, 0L);
         break;


      // Must be some system command -- pass it on to the default
      //  window procedure.

      default:
         return DefWindowProc(hWnd, message, wParam, lParam);
   }

   return (NULL);
}

//****************************************************************************
// Function: FormatText
//
// Purpose: Called whenever there is a need to format all the text and
//          paint the first page.  This is called when the window is first
//          created and then each time the window is resized or a new font
//          is selected.
//
// Parameters:                                  
//    hWnd       == Handle to the top level window.
//    hScrDC     == Handle to DC of the window displaying text.
//    hPrnDC     == Hanlde to DC of the printer.
//
// Global Variables Modified:
//  
// 	  guNumPages, guNumLines, glpPages, glpLines, guNumLinesPerPage, 
//    guFirstVisLine, guLastVisLine
//
// Calls:
//    GetNextToken() to get the next token in the text stream.
//    GotoNewLine() to advance to a new line (new paragraph or wrapped line).
//
// Returns : nothing.
//
// Comments:
//
//     1) A token is defined as a text stream without control characters and
//        delimiters.
//     2) The delimiters are ' ' and '-'.
//
// Algorithm:
//
//     1) Initialize variables to represent a new document.
//     2) Get next token in stream. If none left, we are done.
//     3) If token == end of line (\n), then advance to new line.
//     4) If token == control character, ignore it.
//     5) Check if the token fits on this line when drawn. If yes, goto 8.
//     6) Since there is no space for taken on this line, move to next line.
//     7) Goto  2.
//     8) Draw the current token. Since leading blanks in a wrapped line are
//        not necessary, avoid them.  
//
// History:  Date       Author        Reason
//           1/5/93     Krishna       Created
//****************************************************************************

void FormatText(HWND hWnd, HDC hScrDC, HDC hPrnDC)
{
	TEXTMETRIC tm;
	UINT uCharsRead = 0;  // number of characters read so far
	UINT uLineHt;        // height of a line
	UINT xPos, yPos;
	UINT uStrExt;
	char lpszToken[200];
	HBRUSH hBr;
	HRGN hRgn;
	RECT rc;
	
	SetCursor(LoadCursor(NULL, IDC_WAIT));
	xPos = grc.left;
	yPos = grc.top;
	
  	hBr = CreateSolidBrush(GetBkColor(hScrDC));
	FillRect(hScrDC, (LPRECT)&grc, hBr);
	DeleteObject(hBr);

	// set a clipping region to avoid characters extending beyond their
	// character cells from writing into the left and right margins
	hRgn = CreateRectRgnIndirect((LPRECT)&grc);
	SelectClipRgn(hScrDC, hRgn);
	DeleteObject((HGDIOBJ)hRgn);
     
	guNumPages = guNumLines = glpPages[0].uFirstLine = glpPages[0].uLastLine = 1;
	glpLines[0].bNewPara = TRUE;
    GetTextMetrics(hPrnDC, (LPTEXTMETRIC)&tm);
    uLineHt = tm.tmHeight + tm.tmExternalLeading;
    guNumLinesPerPage = grc.bottom / uLineHt;
    
    GetClientRect(hWnd, (LPRECT)&rc);    
    guNumLinesPerScrPage = (rc.bottom - rc.top) / MulDiv(uLineHt, gnScrXRes, gnPrnXRes);
    
    while (uCharsRead < guBuffLen)
    {
		GetNextToken((LPSTR)(glpszText + uCharsRead), (LPSTR)lpszToken,
    	             (grc.right - grc.left) / tm.tmAveCharWidth);
    	// update the current character position	
    	uCharsRead += lstrlen(lpszToken);
    	    	
		if (lpszToken[0] == '\n') // || lpszToken[0] == '\r')
    	{
			// line is being terminated with a newline, so it has no white space left at end.    	
    		glpLines[guNumLines - 1].uWS = 0;
    		glpLines[guNumLines].uFirstChar = uCharsRead;
    		glpLines[guNumLines].bNewPara = TRUE;
    				
			// initialize variables for new line
			xPos = grc.left;
			yPos += uLineHt;    		
    		if (GotoNewLine() < 0)
    			break;
    		else
    			continue;    	
    	}
    	else if iscntrl(lpszToken[0])
    		continue;    // ignore other control characters; continue with next token.
    	else	
    		uStrExt = LOWORD(GetTextExtent(hPrnDC, lpszToken, lstrlen(lpszToken)));
    		
    	if ((xPos + uStrExt) >	(UINT)grc.right)
    	{
    		glpLines[guNumLines - 1].uWS = grc.right - xPos;
    		glpLines[guNumLines].uFirstChar = uCharsRead - lstrlen(lpszToken);
    				
			// initialize variables for new line
			xPos = grc.left;
			yPos += uLineHt;                    
			// a line is being wrapped, so it's not a new paragraph
    		glpLines[guNumLines].bNewPara = FALSE;			    		    		
    		if (GotoNewLine() < 0)
    			break;
    	}		
    	// check to see if the first token in line is a blank to avoid leading spaces in
    	// wrapped around lines. 	
    	if (!(lpszToken[0] == ' ' && xPos == (UINT)grc.left && !glpLines[guNumLines - 1].bNewPara))
    		xPos += uStrExt;
    }
    guFirstVisLine = glpPages[0].uFirstLine;
    guLastVisLine = glpPages[0].uLastLine;
    
	SetCursor(LoadCursor(NULL, IDC_ARROW));    
}

//****************************************************************************
// Function: GotoNewline
//
// Purpose: Called whenever there is a need to advance to a new line.
//
// Parameters:
//    none.
//
// Global Variables Modified:
//  
// 	  guNumPages, guNumLines, glpPages, glpLines
//
// Returns : negative number if the document cannot accomodate a new line.
//           positive number if the document can accomodate a new line.
//
// Comments:
//    This routine adds a new line if the document can allow a new line (and
//    a new page if the current page is full).    
//
// History:  Date       Author        Reason
//           1/5/93     Krishna       Created
//****************************************************************************

int GotoNewLine(void)
{
    glpPages[guNumPages - 1].uLastLine = guNumLines;
    // go to new line when you find '\n' or '\r'
	if (guNumLines < MAXLINES)
	{                                 
		guNumLines++;
		// check to see if we need to place this new line in the next page
		if ((guNumLines - glpPages[guNumPages - 1].uFirstLine) >= guNumLinesPerPage)
			// need to go to next page
			if (guNumPages < MAXPAGES)
			{
				// one more page can be accommodated
				glpPages[guNumPages].uFirstLine = glpPages[guNumPages].uLastLine = guNumLines;
				guNumPages++;
    					
				// continue with the next token in text stream.
				return 1;    					
			}
			else
				return -1;  // cannot allow more pages in document; return from routine.
		else
		{
			// no need to go to next page
			glpPages[guNumPages - 1].uLastLine++;
    				
			// continue with the next token in text stream.
			return 1;
		}	
	}
	else
		return -1;  // cannot allow more lines in document; return from routine.
}

//****************************************************************************
// Function: GetNextToken
//
// Purpose: Retrieves the next token from the text stream.
//
// Parameters:
//    lpszBuff  == Pointer to stream of text. The first character of this
//                 stream is the first character of the token.
//    lpszToken == Pointer to variable holding the token on return.
//    uMaxToken == Maximum length of the token. Tokens longer than this will
//                 be truncated to this size.
//
// Returns : Nothing.
//
// Comments:
//
//     1) A token is defined as a text stream without control characters and
//        delimiters.
//     2) The delimiters are ' ' and '-'.
//
// History:  Date       Author        Reason
//           1/5/93     Krishna       Created
//****************************************************************************
    			

void GetNextToken(LPSTR lpszBuff, LPSTR lpszToken, UINT uMaxToken)
{
	char ch;
	int i = 0;
	
	ch = lpszBuff[0];
	if (iscntrl(ch) || ch == ' ' || ch == '-')
	{
		lpszToken[0] = ch;
		lpszToken[1] = '\0';
	}
	else 
   {
      do
	   {
		   lpszToken[i] = lpszBuff[i];
		   ch = lpszBuff[++i];
	   }
	   while (!(i > (int)uMaxToken || iscntrl(ch) || ch == ' ' || ch == '-'));
	   lpszToken[i] = '\0';
   }
}

//****************************************************************************
// Function: RepaintClient
//
// Purpose: Called whenever there is a need to refresh the display. 
//
// Parameters:
//    ghScrDC     == Handle to DC of the window displaying text.
//    lprc    == pointer to the text formatting rectangle.
//
// Calls:
//    GetNextToken() to get the next token in the text stream.
//    GotoNewLine() to advance to a new line (new paragraph or wrapped line).
//
// Returns : nothing.
//
// Comments:
//
//     1) It uses the existing format data structures, but does not modify them.
//     2) The algorithm is similar to that of FormatAndPaint with the differnce
//        that this routine does not have to update the format information.
//
// Algorithm:
//
//     1) Get next token in stream. If none left, we are done.
//     2) If token == end of line (\n), proceed to draw next line.
//     3) If token == control character, ignore it.
//     4) Check if the token fits on this line when drawn. If yes, goto 7.
//     5) Since there is no space for token on this line, it must be starting on
//        next line. Proceed to draw the next line.
//     6) Goto  1.
//     7) Draw the current token. Since leading blanks in a wrapped line are
//        not necessary, avoid them.  
//
// History:  Date       Author        Reason
//           1/5/93     Krishna       Created
//****************************************************************************

void RepaintClient(HDC hScrDC, HDC hPrnDC, UINT uTopLine, UINT uBotLine)
{
	UINT xPos, yPos;
	UINT i, uCharsRead, uStrExt, uLineHt;
	char lpszToken[200];                
	TEXTMETRIC tm;
	HBRUSH hBr;
	HRGN hRgn;
	HDC hDC = hPrnDC;	// the DC getting the output. printer DC is the default.
    
    if (hScrDC)
    {
    	// set hDC to screen only if we are writing to screen. if we are writing to printer
    	// then the ghScrDC parameter will be NULL.
    	hDC = hScrDC;
    	hBr = CreateSolidBrush(GetBkColor(hScrDC));
    	FillRect(ghScrDC, (LPRECT)&grc, hBr);
    	DeleteObject(hBr);
    }	

	// set a clipping region to avoid characters extending beyond their
	// character cells (such as some TrueType characters, particularly those that are
	// italic) from writing into the left and right margins
	hRgn = CreateRectRgnIndirect((LPRECT)&grc);
	SelectClipRgn(hDC, hRgn);
	DeleteObject((HGDIOBJ)hRgn);

	GetTextMetrics(hPrnDC, (LPTEXTMETRIC)&tm);
	uLineHt = tm.tmHeight + tm.tmExternalLeading;    
    xPos = grc.left;
    yPos = grc.top; 
    uCharsRead = glpLines[uTopLine - 1].uFirstChar;
	for (i = uTopLine; i <= uBotLine; uCharsRead = glpLines[i++].uFirstChar)
    	while (uCharsRead  < guBuffLen)
    	{
			GetNextToken((LPSTR)(glpszText + uCharsRead), (LPSTR)lpszToken, 
			             (grc.right - grc.left) / tm.tmAveCharWidth);
    		// update the current character position	
    		uCharsRead += lstrlen(lpszToken);
    	    	
			if (lpszToken[0] == '\n') // || lpszToken[0] == '\r')
			{
				yPos += uLineHt;
				xPos = grc.left;
				break;
			}	
    		else if iscntrl(lpszToken[0])
    			continue;    // ignore other control characters; continue with next token.
    		else	
    			uStrExt = LOWORD(GetTextExtent(hPrnDC, lpszToken, lstrlen(lpszToken)));
    		
    		if ((xPos + uStrExt) >	(UINT)grc.right)
    		{
				// initialize variables for new line
				yPos += uLineHt;
				xPos = grc.left;
				break;
			}
    		else if (!(lpszToken[0] == ' ' && xPos == (UINT)grc.left && !glpLines[i - 1].bNewPara))
			{	    		    		
    			ExtTextOut(hDC, xPos, yPos, NULL, NULL, lpszToken, lstrlen(lpszToken), NULL);
    			xPos += uStrExt;
    		}
    			
		}
}

//****************************************************************************
// Function: Scroll
//
// Purpose: Called whenever the text has to be scrolled.
//
// Parameters:
//    hWnd == Handle to window whose client area has text to scroll.
//
// Global Variables Modified:
//  
// 	  guFirstVisLine, guLastVisLine
//
// Calls:
//    RepaintClient() to repaint the scrolled text.
//
// Returns : nothing.
//
// Comments:
//
//     1) This is a primitive scrolling routine. It uses RepaintClient() to 
//        paint the text and as a result has a flash (due to the FillRect 
//        used to erase previous text). This flash can be avoided by using 
//        ScrollDC or ScrollWindow and repainting the newly uncovered text. 
//
// History:  Date       Author        Reason
//           1/5/93     Krishna       Created
//****************************************************************************

void Scroll(HWND hWnd)
{
	if ((int)guFirstVisLine < 1)
	{
		guFirstVisLine = 1;
		guLastVisLine = guNumLinesPerScrPage;
	}
	guLastVisLine = guFirstVisLine + guNumLinesPerScrPage - 1;	
	if (guLastVisLine > guNumLines)
		guLastVisLine = guNumLines;
          
	RepaintClient(ghScrDC, ghPrnDC, guFirstVisLine, guLastVisLine);	
	SetScrollPos(hWnd, SB_VERT, MulDiv(guFirstVisLine, 10000, guNumLines), TRUE);
}	
